/*
 * RandomTreeFactory.cpp
 *
 *  Created on: 18. 3. 2015
 *	  Author: Stepan Plachy
 */

#include "RandomTreeFactory.h"

#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <cstdlib>
#include <ctime>
#include <cmath>

#include <iostream>

namespace tree {

namespace generate {

int c = 0;


struct Node {
	char symbol;
	int depth;
	Node * right;
	Node * child = NULL;
	int rank = 0;

	Node() : depth(0) {}
	Node(Node * parent) : depth(parent -> depth + 1) {
		if (parent -> child == NULL) {
			parent -> child = this;
			right = this;
		} else {
			right = parent -> child -> right;
			parent -> child -> right = this;
		}
		parent -> rank++;
	}
	~Node() {
		if (child == NULL) return;
		Node * ch = child;
		do {
			Node * tmp = ch -> right;
			delete ch;
			ch = tmp;
		} while (ch != child);
	}

	void rotateLeftBranch() {
		if (child != NULL) child -> rotateLeftBranch();
		if (rank > 1) {
			int angle = rand() % rank;
			Node * newChild = child;
			for (int i = 0; i < angle; i++) {
				newChild = newChild -> right;
			}
			child = newChild;
		}
	}

	void generateUnrankedSymbols(const std::vector<char> & alphabet) {
		symbol = alphabet[rand() % alphabet.size()];
		Node * nextChild = child;
		for(int i = 0; i < rank; i++) {
				nextChild -> generateUnrankedSymbols(alphabet);
				nextChild = nextChild -> right;
		}
	}

	void generateRankedSymbols(const std::map<int, std::vector<char> > rankedAlphabet) {
		const std::vector<char> & alphabet = rankedAlphabet.at(rank);
		symbol = alphabet[rand() % alphabet.size()];
		Node * nextChild = child;
		for(int i = 0; i < rank; i++) {
				nextChild -> generateRankedSymbols(rankedAlphabet);
				nextChild = nextChild -> right;
		}
	}

	void fillRanks(std::map<int, std::vector<char> > & rankedAlphabet) {
		rankedAlphabet[rank];
		Node * nextChild = child;
		for(int i = 0; i < rank; i++) {
				nextChild -> fillRanks(rankedAlphabet);
				nextChild = nextChild -> right;
		}
	}

	UnrankedNode * createUnrankedNode() {
		std::vector<UnrankedNode *> children;
		Node * nextChild = child;
		for(int i = 0; i < rank; i++) {
				children.push_back(nextChild -> createUnrankedNode());
				nextChild = nextChild -> right;
		}
		return new UnrankedNode(alphabet::LabeledSymbol(symbol), children);
	}

	RankedNode * createRankedNode() {
		std::vector<RankedNode *> children;
		Node * nextChild = child;
		for(int i = 0; i < rank; i++) {
				children.push_back(nextChild -> createRankedNode());
				nextChild = nextChild -> right;
		}
		return new RankedNode(alphabet::RankedSymbol(symbol, rank), children);
	}

	void nicePrint(std::ostream & os = std::cout, const std::string & prefix = "", const bool last = true) const {
		os << prefix;

		std::string nextPrefix(prefix);
		if (last) {
			os << "\\-";
			nextPrefix += "  ";
		} else {
			os << "|-";
			nextPrefix += "| ";
		}
		os << symbol << " (" << rank << ")" << std::endl;

		Node * nextChild = child;
		for (int i = 0; i < rank; i++)
		{
			//os << nextPrefix << "|" << std::endl;
			nextChild -> nicePrint(os, nextPrefix, i == rank-1);
			nextChild = nextChild -> right;
		}
	}
};

std::vector<char> generateUnrankedAlphabet(int maxAlphabetSize) {
	std::vector<char> symbols (26);
	for(int i = 0; i < 26; i++) symbols[i] = i + 'a';
	random_shuffle(symbols.begin(), symbols.end());
	return std::vector<char> (symbols.begin(), symbols.begin() + maxAlphabetSize);
}

void generateRankedAlphabet(std::map<int, std::vector<char> > & rankedAlphabet, int maxAlphabetSize) {
	int ranksCount = rankedAlphabet.size();
	std::vector<char> unrankedAlphabet = generateUnrankedAlphabet(maxAlphabetSize > ranksCount ? maxAlphabetSize : ranksCount);

	std::set<int> rankSeparators;
	rankSeparators.insert(0);
	rankSeparators.insert(unrankedAlphabet.size());
	while ((int)rankSeparators.size() != ranksCount + 1 /*&& rankSeparators.size() != maxRank + 2*/) rankSeparators.insert(rand() % unrankedAlphabet.size());

	std::set<int>::iterator it = rankSeparators.begin();
	for (std::map<int, std::vector<char> >::iterator i = rankedAlphabet.begin(); i != rankedAlphabet.end(); ++i) {
		std::set<int>::iterator prevIt = it++;
		i -> second.insert(i -> second.begin(), unrankedAlphabet.begin() + *prevIt, unrankedAlphabet.begin() + *it);
	}
}

Node * generateTreeStructure(int depth, int nodesCount, int maxRank = INT_MAX) {
	if(depth >= nodesCount) throw exception::AlibException("number of nodes is too small");

	if(maxRank != INT_MAX && pow(maxRank, depth + 1) - 1 < nodesCount) throw exception::AlibException("number of nodes is too small");
	srand(time(NULL));
	std::vector<Node *> nodes (nodesCount);

	//generate path depth long
	Node * root = nodes[0] = new Node();
	for (int i = 1; i <= depth; i++) nodes[i] = new Node(nodes[i-1]);
	//move final leaf to end
	nodes[nodesCount-1] = nodes[depth];
	
	int availableNodesIndex = depth;
	int finalNodesIndex = nodesCount - 2;
	while(finalNodesIndex >= availableNodesIndex) {
		int randomIndex = rand() % availableNodesIndex;
		Node * parent = nodes[randomIndex];
		Node * node = new Node(parent);

		//put node to end if it reached depth limit
		if (node -> depth < depth) {
			nodes[availableNodesIndex] = node;
			availableNodesIndex++;
		} else {
			nodes[finalNodesIndex] = node;
			finalNodesIndex--;
		}

		//put parent node to end if it reached rank limit
		if (parent -> rank >= maxRank) {
			nodes[randomIndex] = nodes[availableNodesIndex-1];
			nodes[finalNodesIndex] = parent;
			availableNodesIndex--;
			finalNodesIndex--;
		}
	}

	//move generated path to random branch
	root -> rotateLeftBranch();

	return root;
}

UnrankedTree RandomTreeFactory::generateUnrankedTree(int depth, int nodesCount, int maxAlphabetSize, int maxRank) {
	Node * root = generateTreeStructure(depth, nodesCount, maxRank);
	std::vector<char> alphabet = generateUnrankedAlphabet(maxAlphabetSize);
	root -> generateUnrankedSymbols(alphabet);

	std::set<alphabet::LabeledSymbol> treeAlphabet;
	for (std::vector<char>::iterator it = alphabet.begin(); it != alphabet.end(); ++it) {
		treeAlphabet.insert(alphabet::LabeledSymbol(*it));
	}
	UnrankedTree tree (treeAlphabet, *std::move(root -> createUnrankedNode()));
	delete root;
	return tree;
}

RankedTree RandomTreeFactory::generateRankedTree(int depth, int nodesCount, int maxAlphabetSize, int maxRank) {
	Node * root = generateTreeStructure(depth, nodesCount, maxRank);
	std::map<int, std::vector<char> > rankedAlphabet;
	root -> fillRanks(rankedAlphabet);
	generateRankedAlphabet(rankedAlphabet, maxAlphabetSize);
	root -> generateRankedSymbols(rankedAlphabet);

	std::set<alphabet::RankedSymbol> treeRankedAlphabet;
	for (std::map<int, std::vector<char> >::iterator it = rankedAlphabet.begin(); it != rankedAlphabet.end(); ++it) {
		std::vector<char> & alphabet = it -> second;
		for (std::vector<char>::iterator i = alphabet.begin(); i != alphabet.end(); ++i) {
			treeRankedAlphabet.insert(alphabet::RankedSymbol(*i, it -> first));
		}
	}
	RankedTree tree (treeRankedAlphabet, std::move(*root -> createRankedNode()));
	delete root;
	return tree;
}

} /* namespace generate */

} /* namespace automaton */
